use super::*;
use std::collections::HashMap;
use tokio::time::{
    DelayQueue,
    delay_queue,
};


#[derive(Debug)]
pub struct Cache
{
    //TODO: HashMap and DelayQueue of mapping public keys to user IDs, etc.
    pkey_maps: HashMap<String, (UserID, delay_queue::Key)>,
    pkey_rems: DelayQueue<String>,
}

impl Cache
{
    /// Create a new empty cache
    pub fn new() -> Self
    {
	Self {
	    pkey_maps: HashMap::new(),
	    pkey_rems: DelayQueue::new(),
	}
    }

    //how tf do we get this to run concurrently with insertions??? it holds `&mut self` forever!
    //redesign required. maybe using Arc and RwLock or Mutex and interrupting the purge task when something needs to be inserted. i.e: give this task a stream that it can `select!` along with calling `next()`, if the other future completes first, we return? but wouldn't that lose us an item in `next()`? is there a `select with priority` in `futures`? i think there is. eh...
    /// Run a purge on this cache.
    pub async fn purge(&mut self) -> eyre::Result<()>
    {
	let mut errors = Vec::default();
	while let Some(entry) = self.pkey_rems.next().await
	{
	    match entry {
		Ok(entry) => {
		    self.pkey_maps.remove(entry.get_ref());
		},
		Err(err) => {
		    errors.push(err);
		}
	    }
	}
	if errors.len() > 0 {
	    let mut err = Err(eyre!("One or more removals failed"))
		.with_note(|| errors.len().to_string().header("Number of failed removals"));
	    for e in errors.into_iter()
	    {
		err = err.with_error(move || e);
	    }
	    return err;
	} else {
	    Ok(())
	}
    }
    /// Purge all values available to be purged now.
    pub fn purge_now(&mut self) -> eyre::Result<()>
    {
	match self.purge().now_or_never() {
	    Some(x) => x,
	    None => Ok(())
	}
    }
}
